import fastapi
from fastapi import Depends, HTTPException, Response
from auth import auth
import logging
import conf
import json
from starlette_context import context
from bundles import bundle
import correlation
from pydantic import BaseModel

from starlette.status import (
    HTTP_201_CREATED,
    HTTP_202_ACCEPTED,
    HTTP_400_BAD_REQUEST,
    HTTP_405_METHOD_NOT_ALLOWED,
    HTTP_424_FAILED_DEPENDENCY,
)


class Detail(BaseModel):
    detail: str


router = fastapi.APIRouter()

logger = logging.getLogger(__name__)

description = f"""
## Experimental bootstrap API for creating and updating bundle to default.

This should be used when adding a partition to OSDU or when wanting to reset to a known good state.
    
### Without force:

- This method is only allowed if the partition doesn't already have a bundle.
- If the bundle already exists it will return 405 METHOD_NOT_ALLOWED.
- Policy Service can be configured to ignore force.

### May return:

- `HTTP_202_ACCEPTED` - updated
- `HTTP_201_CREATED` - created
- `HTTP_405_METHOD_NOT_ALLOWED` - not allowed
- `HTTP_424_FAILED_DEPENDENCY` - bundle server caused failure

### Permission required to use this API on this system (configurable):

    {conf.OPS_PERMISSION}
    """


@router.post(
    "/bootstrap",
    responses={
        HTTP_201_CREATED: {"model": Detail},
        HTTP_202_ACCEPTED: {"model": Detail},
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_405_METHOD_NOT_ALLOWED: {"model": Detail},
    },
    description=description,
)
def bootstrap(
    response: Response,
    auth_data: auth.Auth = Depends(auth.require_authorized_ops),
    force: bool = False,
):
    """
    Experimental bootstrap API for creating and updating bundle to default.
    """

    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)

    response.headers["X-Correlation-ID"] = context["correlation_id"]

    if bundle.get_bundle(data_partition=auth_data.data_partition_id):
        logger.debug("bootstrap bundle already exists")
        if force:
            if not conf.ALLOW_FORCE_BOOTSTRAP:
                raise HTTPException(
                    status_code=HTTP_405_METHOD_NOT_ALLOWED,
                    detail=f"'{auth_data.data_partition_id}' already bootstrapped and forced updates not allowed.",
                )
            status = bootstrap_bundle(data_partition=auth_data.data_partition_id)
            if status:
                response.status_code = HTTP_202_ACCEPTED
                detail = f"bootstrap bundle for partition '{auth_data.data_partition_id}' updated to default"
                audit = f"AUDIT {detail}, user: {auth_data.user_id}"
                logger.info(audit)
                return {"detail": detail}
            else:
                raise HTTPException(
                    status_code=HTTP_424_FAILED_DEPENDENCY,
                    detail=f"'{auth_data.data_partition_id}' update failed {status}!",
                )
        else:
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail=f"'{auth_data.data_partition_id}' already bootstrapped.",
            )
    else:
        status = bootstrap_bundle(data_partition=auth_data.data_partition_id)
        if status:
            response.status_code = HTTP_201_CREATED
            detail = f"bootstrap bundle for partition '{auth_data.data_partition_id}' created"
            audit = f"AUDIT {detail}, user: {auth_data.user_id}"
            logger.info(audit)
            return {"detail": detail}
        else:
            detail = f"bootstrap bundle for partition '{auth_data.data_partition_id}' create failed {status}!"
            logger.error(detail)
            raise HTTPException(status_code=HTTP_424_FAILED_DEPENDENCY, detail=detail)


def bootstrap_bundle(data_partition):
    # default dataauthz rego for a partition
    dataauthz = """\
package osdu.partition["{}"].dataauthz

import data.osdu.instance.dataauthz as centralauthz

records := centralauthz.records""".format(data_partition)

    # default search rego for a partition
    search = """\
package osdu.partition["{}"].search

default allow = false

allow = true {{
    input.operation == "view"
    # At least one user group needs to be in acl viewers
    input.record.acl.viewers[_]==input.groups[_]
}}

allow = true {{
    input.operation == ["view", "create", "update", "delete", "purge"][_]
    # At least one user group needs to be in acl owners
    input.record.acl.owners[_]==input.groups[_]
}}""".format(data_partition)

    # manifest is json
    manifest = {"roots": [f"osdu/partition/{data_partition}"]}

    #    manifest = """\
    # {{
    #    "roots": ["osdu/partition/{}"]
    # }}""".format(data_partition)

    default_policy_list = {
        # ".manifest": manifest.encode('ascii'),
        ".manifest": json.dumps(manifest, indent=0).encode("ascii"),
        "dataauthz.rego": dataauthz.encode("ascii"),
        "search.rego": search.encode("ascii"),
    }

    return bundle.put_default_bundle(
        data_partition=data_partition, policy_list=default_policy_list
    )
