from fastapi import APIRouter, Depends, UploadFile, HTTPException, Query, Response
from starlette_context import context
import logging
from auth import auth
import opa
import correlation
from pydantic import BaseModel

from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_422_UNPROCESSABLE_ENTITY,
)


class Detail(BaseModel):
    detail: str


router = APIRouter()
logger = logging.getLogger(__name__)


@router.post(
    "/compile",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_422_UNPROCESSABLE_ENTITY: {"model": Detail},
    },
)
def compile_partially_evaluate_a_query(
    response: Response,
    file: UploadFile,
    metrics: bool = Query(
        description="Include report detailed performance metrics on requested on individual API call. Returned inline with the API response",
        default=False,
    ),
    instrument: bool = Query(
        description="Include instrumentation data wth detailed performance metrics on requested on individual API call. Returned inline with the API response",
        default=False,
    ),
    auth_data: auth.Auth = Depends(auth.require_authorized_admin),
):
    """
    # Compile - Partially evaluate a query.
    The Compile API allows you to partially evaluate Rego queries and obtain a simplified version of the policy.

    ### Metrics
    When query parameter metrics=true, the API response will include detailed performance metrics from OPA.
    OPA currently supports the following query performance metrics:

        timer_rego_input_parse_ns: time taken (in nanoseconds) to parse the input
        timer_rego_query_parse_ns: time taken (in nanonseconds) to parse the query.
        timer_rego_query_compile_ns: time taken (in nanonseconds) to compile the query.
        timer_rego_query_eval_ns: time taken (in nanonseconds) to evaluate the query.
        timer_rego_module_parse_ns: time taken (in nanoseconds) to parse the input policy module.
        timer_rego_module_compile_ns: time taken (in nanoseconds) to compile the loaded policy modules.
        timer_server_handler_ns: time take (in nanoseconds) to handle the API request.

    ### Instrumentation
    To enable query instrumentation, specify metrics=true and instrument=true query parameters when executing the API call.
    Query instrumentation can help diagnose performance problems, however, it can add significant overhead to query evaluation.
    We recommend leaving query instrumentation off unless you are debugging a performance problem.

    ### Permission required to use this API:

        policy.service.admin
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    # contents = await file.read()
    contents = file.file.read()

    try:
        data = contents.decode("utf-8")
    except UnicodeDecodeError as err:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail=f"decode error {err}"
        )

    opa_response = opa.compile(query=data, metrics=metrics, instrument=instrument)
    logger.info(f"OPA compile return {opa_response.json}")

    if opa_response.ok:
        return opa_response.json
    else:
        raise HTTPException(
            status_code=opa_response.status_code, detail=opa_response.message
        )
