# policy_api.py

# Copyright 2022 <XYZ>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fastapi
from fastapi import Header, Depends, HTTPException, Response
from typing import Union
import sys
import os
import logging
from starlette_context import context
import hashlib
from pydantic import BaseModel

from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

from opa_response import OpaResponse

sys.path.append(os.path.abspath(".."))
import conf
import opa
from auth import auth
import correlation


class Detail(BaseModel):
    detail: str


logger = logging.getLogger(__name__)
router = fastapi.APIRouter()

# Header code is compatible with Python 3.6 - 3.9, it will not work with Python 3.10
# See https://fastapi.tiangolo.com/tutorial/header-params/


@router.get(
    "/policies",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_500_INTERNAL_SERVER_ERROR: {"model": Detail},
    },
)
def fetch_all_policies_that_match_partition(
    response: Response,
    data_partition_id: Union[str, None] = Header(default=None),
    auth_data: auth.Auth = Depends(auth.require_authorized_user),
):
    """
    Return all policies from OPA directly that match partition_bundle_root data-partition-id in header (if bundles are enabled).
    This API gives the list of all the defined policies and it includes the policy definitions in the raw Rego form.
    It performs authorization check. The user making the call needs to be either service.policy.user or service.policy.admin in the provided data partition.

    ### Permission required to use this API:

    `policy.service.user` or `policy.service.admin`
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]
    logging.debug(f"data_partition_id: {data_partition_id}")

    cloud_provider = os.environ.get("CLOUD_PROVIDER")
    if cloud_provider is None:
        opa_response = OpaResponse()
        logger.critical(
            "Error: CLOUD_PROVIDER ENV VAR not set / Mocking results for /policies"
        )
        opa_response.json = {"result": {}}
    elif cloud_provider == conf.MOCK:
        opa_response = OpaResponse()
        logger.critical(
            "Warning: CLOUD_PROVIDER ENV VAR set to Mock results for /policies"
        )
        opa_response.json = {"result": {}}
    else:
        opa_response = opa.fetch_opa_policies_direct()
        if not opa_response.ok:
            raise HTTPException(
                status_code=opa_response.status_code,
                detail=f"Unexpected result from OPA {opa_response.message}",
            )

    if conf.USE_BUNDLES and "result" in opa_response.json:
        logger.debug(f"OPA return {opa_response.json}")
        policies = opa_response.json
        # we need to filter and return only the instance policies and policies for requested data partition id
        return {
            "result": [
                x
                for x in policies["result"]
                if x["id"].startswith(
                    conf.PARTITION_BUNDLE_ROOT.format(data_partition_id)
                )
                or x["id"].startswith(
                    "/" + conf.PARTITION_BUNDLE_ROOT.format(data_partition_id)
                )
                or x["id"].startswith(conf.INSTANCE_BUNDLE_ROOT)
                or x["id"].startswith("/" + conf.INSTANCE_BUNDLE_ROOT)
            ]
        }
    raise HTTPException(
        status_code=HTTP_500_INTERNAL_SERVER_ERROR, detail="Unexpected result from OPA"
    )


@router.get(
    "/policies/{policy_id}",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_404_NOT_FOUND: {"model": Detail},
        HTTP_500_INTERNAL_SERVER_ERROR: {"model": Detail},
    },
)
def fetch_a_policy(
    response: Response,
    policy_id: str,
    auth_data: auth.Auth = Depends(auth.require_authorized_user),
):
    """
    Return a policy directly from OPA with no filtering

    ### Permission required to use this API:

    `policy.service.user` or `policy.service.admin`
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    cloud_provider = os.environ.get("CLOUD_PROVIDER")
    if cloud_provider is None:
        # Help support MOCK testing
        opa_response = OpaResponse()
        logger.critical(
            f"Error: CLOUD_PROVIDER ENV VAR not set / Mocking results for /policies/{policy_id}"
        )
        opa_response.message = f"MOCK Policy {policy_id} added to OPA"
        opa_response.ok = False
        opa_response.json = {"result": {}}
        opa_response.status_code = HTTP_404_NOT_FOUND
    elif cloud_provider == conf.MOCK:
        opa_response = OpaResponse()
        logger.critical(
            f"Warning: CLOUD_PROVIDER ENV VAR set to Mock results {cloud_provider} for /policies/{policy_id}"
        )
        opa_response.message = f"MOCK Policy {policy_id} added to OPA"
        opa_response.ok = False
        opa_response.json = {"result": {}}
        opa_response.status_code = HTTP_404_NOT_FOUND
    else:
        opa_response = opa.fetch_opa_policy_direct(policy_id)

    if opa_response.ok is True:
        return opa_response.json
    raise HTTPException(
        status_code=opa_response.status_code,
        detail=f"Unexpected result from OPA {opa_response.message}",
    )


@router.get(
    "/policies/osdu/instance/{policy_id}",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_500_INTERNAL_SERVER_ERROR: {"model": Detail},
    },
)
def fetch_instance_policy(
    response: Response,
    policy_id: str,
    auth_data: auth.Auth = Depends(auth.require_authorized_user),
):
    """
    Return an instance policy from OPA directly.

    ### Permission required to use this API:

    `policy.service.user` or `policy.service.admin`
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    cloud_provider = os.environ.get("CLOUD_PROVIDER")
    if cloud_provider is None:
        # Help support MOCK testing
        opa_response = OpaResponse()
        logger.critical(
            f"Error: CLOUD_PROVIDER ENV VAR not set / Mocking results for /policies/osdu/instance/{policy_id}"
        )
        opa_response.message = f"MOCK OK policy_id: {policy_id}"
        opa_response.ok = True
        opa_response.json = {"result": {}}
    elif cloud_provider == conf.MOCK:
        opa_response = OpaResponse()
        logger.critical(
            f"Warning: CLOUD_PROVIDER ENV VAR set to Mock results for /policies/osdu/instance/{policy_id}"
        )
        opa_response.message = f"MOCK OK policy_id: {policy_id}"
        opa_response.ok = True
        opa_response.json = {"result": {}}
    else:
        search = conf.INSTANCE_BUNDLE_ROOT + "/{0}".format(policy_id)
        logger.debug(f"fetching instance profile {search}")
        opa_response = opa.fetch_opa_policy_direct(
            conf.INSTANCE_BUNDLE_ROOT + "/{0}".format(policy_id)
        )

    logger.debug(f"OPA return {opa_response.json}")
    if opa_response.ok is True:
        return opa_response.json
    raise HTTPException(
        status_code=opa_response.status_code,
        detail=f"Unexpected result from OPA {opa_response.message}",
    )


@router.get(
    "/policies/osdu/partition/{data_partition}/{policy_id}",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_500_INTERNAL_SERVER_ERROR: {"model": Detail},
    },
)
def fetch_partition_policy_directly_from_opa(
    response: Response,
    policy_id: str,
    data_partition: str,
    auth_data: auth.Auth = Depends(auth.require_authorized_user),
):
    """
    Return an policy for a partition id from OPA.
    Requires data-partition-id in header.

    ### Permission required to use this API:

    `policy.service.user` or `policy.service.admin`
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    if data_partition != auth_data.data_partition_id:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Provided data partition id does not correspond to provided policy id",
        )

    cloud_provider = os.environ.get("CLOUD_PROVIDER")
    if cloud_provider is None:
        # Help support MOCK testing
        opa_response = OpaResponse()
        logger.critical(
            f"Error: CLOUD_PROVIDER ENV VAR not set / Mocking results for /policies/osdu/partition/{data_partition}/{policy_id}"
        )
        opa_response.message = f"MOCK OK policy_id: {policy_id}"
        opa_response.json = {"result": {}}
        opa_response.ok = True
    elif cloud_provider == conf.MOCK:
        opa_response = OpaResponse()
        logger.critical(
            f"Warning: CLOUD_PROVIDER ENV VAR set to Mock results for /policies/osdu/partition/{data_partition}/{policy_id}"
        )
        opa_response.message = f"MOCK OK policy_id: {policy_id}"
        opa_response.json = {"result": {}}
        opa_response.ok = True
    else:
        # Force slash '/' into the URL, seems like an OPA bug
        if conf.OPA_BUG_FORCE_SLASH:
            pid = (
                "%2f"
                + conf.PARTITION_BUNDLE_ROOT.format(data_partition)
                + "/{0}".format(policy_id)
            )
        else:
            pid = conf.PARTITION_BUNDLE_ROOT.format(data_partition) + "/{0}".format(
                policy_id
            )
        logger.debug(f"PID = {pid}")
        opa_response = opa.fetch_opa_policy_direct(pid)

    logger.debug(f"OPA return {opa_response.json}")

    if conf.ENABLE_DEV_DIAGNOSTICS:
        # include extra debugging information in response headers
        response.headers["X-Debug-Result"] = f"{opa_response.json}"

    if opa_response.ok is True:
        if "result" in opa_response.json:
            if "raw" in opa_response.json["result"]:
                data = opa_response.json["result"]["raw"]
                sha1 = hashlib.sha1(data.encode()).hexdigest()
                response.headers["X-SHA-1"] = sha1

        return opa_response.json

    raise HTTPException(
        status_code=opa_response.status_code,
        detail=f"Unexpected result from OPA {opa_response.message}",
    )
