# policy_api.py

# Copyright 2022 <XYZ>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fastapi
from fastapi import Depends, HTTPException, UploadFile, Response
import sys
import os
import logging
import hashlib
from starlette.requests import Request
from starlette_context import context
# from requests.structures import CaseInsensitiveDict

from starlette.status import (
    HTTP_200_OK,
    HTTP_202_ACCEPTED,
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
    HTTP_501_NOT_IMPLEMENTED,
    HTTP_503_SERVICE_UNAVAILABLE,
)

from pydantic import BaseModel
from opa_response import OpaResponse

sys.path.append(os.path.abspath(".."))
import conf
import opa
from auth import auth
from bundles import bundle
import correlation
from .validate_api import verify_policy_with_opa

logger = logging.getLogger(__name__)
router = fastapi.APIRouter()


class Detail(BaseModel):
    detail: str


# Header code is compatible with Python 3.6 - 3.9, it will not work with Python 3.10
# See https://fastapi.tiangolo.com/tutorial/header-params/


@router.delete(
    "/policies/osdu/partition/{data_partition}/{policy_id}",
    responses={
        HTTP_202_ACCEPTED: {"model": Detail},
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_404_NOT_FOUND: {"model": Detail},
        HTTP_501_NOT_IMPLEMENTED: {"model": Detail},
        HTTP_503_SERVICE_UNAVAILABLE: {"model": Detail},
    },
)
def delete_partition_policy(
    response: Response,
    request: Request,
    policy_id: str,
    data_partition: str,
    auth_data: auth.Auth = Depends(auth.require_authorized_admin),
):
    """
    ## Delete a partition policy
    * This API requires admin privileges (service.policy.admin) in the provided data partition.
    * Partition ID in header and in path must match
    * Policy_id must end with a ".rego"

    ### Possible http return status codes:
    * 200 - no error
    * 202 - accepted
    * 400 - bad request
        * for example, if data_partition in path doesn't match data-partition-id in header
    * 401 - unauthorized
    * 403 - forbidden
        * for example, if calling with only user privs
    * 404 - not found
    * 422 - validation Error
        * for example, if policy_id doesn't end with ".rego"
    * 500 - server error
    * 501 - not implemented
        * for example, if bundles are not supported.
    * 503 - service not available
        * for example, if issues with bundles server

    Errors will include some detail in returning json.

    Return json:
    ```
        {
            "policy_id": string,
            "data_partition": string,
            "status": bool,
            "message": string,
            "result": string json from OPA
        }
    ```

    ### Permission required to use this API:

    `policy.service.admin`
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    # If uvicorn is running with proxy headers enabled (--proxy-headers)
    x_forwarded_for = request.headers.get("x-forwarded-for")

    # x-request-id
    x_request_id = request.headers.get("x-request-id")

    # Zipkin
    # B3 Propagation is a specification for the header "b3" and those that start with "x-b3-".
    # These headers are used for trace context propagation across service boundaries.
    x_b3_traceid = request.headers.get("x-b3-traceid")
    x_b3_spanid = request.headers.get("x-b3-spanid")
    x_b3_parentspanid = request.headers.get("x-b3-parentspanid")

    audit_msg = f"user: {auth_data.user_id} xuserid: {auth_data.x_user_id} requestid: {x_request_id} traceid: {x_b3_traceid} spanid: {x_b3_spanid} parentspanid: {x_b3_parentspanid} forward: {x_forwarded_for}"

    if conf.USE_BUNDLES is None:
        raise HTTPException(
            status_code=HTTP_501_NOT_IMPLEMENTED,
            detail="Partition specific policies are not supported on this OSDU instance",
        )

    if data_partition != auth_data.data_partition_id:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Provided data partition id does not correspond to provided policy id",
        )

    if not policy_id.endswith(".rego"):
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Policy ID should end with a .rego",
        )

    cloud_provider = os.environ.get("CLOUD_PROVIDER")
    id = f"osdu/partition/{data_partition}/{policy_id}"
    if cloud_provider is None:
        # Help support MOCK testing
        result = OpaResponse()
        logger.critical(
            f"Error: CLOUD_PROVIDER ENV VAR not set / Mocking results for delete: /policies/osdu/partition/{data_partition}/{id}"
        )
        result.message = f"MOCK OK policy_id: {id}"
        result.ok = True
    elif cloud_provider == conf.MOCK:
        result = OpaResponse()
        logger.critical(
            f"Warning: CLOUD_PROVIDER ENV VAR set to Mock results {cloud_provider} for delete: /policies/osdu/partition/{data_partition}/{id}"
        )
        result.message = f"MOCK OK policy_id: {id}"
        result.ok = True
    elif cloud_provider == conf.LOCAL:
        # LOCAL Development ENV
        logger.critical(
            f"Error: CLOUD_PROVIDER ENV VAR LOCAL, Using local / direct OPA {conf.OPA_URL} delete /policies/osdu/partition/{data_partition}/{id}"
        )
        result = opa.delete_opa_policy_direct(id)
    else:
        logger.info(
            f"bundle delete policy_id: {policy_id} id: {id} {auth_data.user_id}"
        )
        result = bundle.delete_policy(
            data_partition=data_partition, policy_id=policy_id
        )

    if result.ok:
        logger.info(f"AUDIT delete policy_id: {policy_id} id: {id} {audit_msg}")
        return {
            "policy_id": policy_id,
            "data_partition": data_partition,
            "status": result.ok,
            "message": result.message,
            "result": result.json,
        }
    else:
        if cloud_provider == conf.LOCAL:
            logger.error(f"Error when talking to OPA {result.message}")
            raise HTTPException(
                status_code=result.status_code,
                detail=f"Error when talking to OPA {result.message}",
            )
        else:
            logger.error(f"Error when talking to bundle service {result.message}")
            raise HTTPException(
                status_code=result.status_code,
                detail=f"Error when talking to bundle service {result.message}",
            )


@router.put(
    "/policies/osdu/partition/{data_partition}/{policy_id}",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_404_NOT_FOUND: {"model": Detail},
        HTTP_500_INTERNAL_SERVER_ERROR: {"model": Detail},
        HTTP_501_NOT_IMPLEMENTED: {"model": Detail},
        HTTP_503_SERVICE_UNAVAILABLE: {"model": Detail},
    },
)
def create_or_update_partition_policy(
    response: Response,
    request: Request,
    policy_id: str,
    data_partition: str,
    file: UploadFile,
    auth_data: auth.Auth = Depends(auth.require_authorized_admin),
):
    """

    ## Create or update a policy with given policy_id for a given data partition.
    This API will create/update policy definition with provided Rego expression included in file and assign it with provided id.

    * This API requires admin privileges (service.policy.admin) in the provided data partition.
    * Partition ID in header and in path must match
    * Policy_id must end with a ".rego"

    ### Possible http return status codes:
    * 200 - no error
    * 202 - accepted
    * 400 - bad request
        * for example, if data_partition in path doesn't match data-partition-id in header
    * 401 - unauthorized
    * 403 - forbidden
        * for example, if calling with only user privs
    * 422 - validation Error
        * for example, if policy_id doesn't end with ".rego"
        * for example, if package declaration issue
    * 500 - server error
    * 501 - not implemented
        * for example, if bundles are not supported.
    * 503 - service not available
        * for example, if issues with bundles server

    Errors will include some detail in returning json.

    Return json:
    ```
        {
            "policy_id": string,
            "data_partition": string,
            "opa_payload": string,
            "status_code": http status code
            "status": bool
            "message": string
        }
    ```

    ### Permission required to use this API:

    `policy.service.admin`
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    # If uvicorn is running with proxy headers enabled (--proxy-headers)
    x_forwarded_for = request.headers.get("x-forwarded-for")

    # x-request-id
    x_request_id = request.headers.get("x-request-id")

    # B3 Propagation is a specification for the header "b3" and those that start with "x-b3-".
    # These headers are used for trace context propagation across service boundaries.
    # Most commonly used by Zipkin
    x_b3_traceid = request.headers.get("x-b3-traceid")
    x_b3_spanid = request.headers.get("x-b3-spanid")
    x_b3_parentspanid = request.headers.get("x-b3-parentspanid")

    if conf.USE_BUNDLES is None:
        raise HTTPException(
            status_code=HTTP_501_NOT_IMPLEMENTED,
            detail="Partition specific policies are not supported on this OSDU instance",
        )

    if auth_data.data_partition_id != data_partition:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Provided data partition id does not correspond to provided policy id",
        )

    if not policy_id.endswith(".rego"):
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Policy ID should end with a .rego",
        )

    # contents = await file.read()
    contents = file.file.read()

    if not bundle.package_name_ok(data_partition, policy_id, contents):
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Package declaration should be: {0}".format(
                bundle.correct_package_declaration(data_partition, policy_id)
            ),
        )

    cloud_provider = os.environ.get("CLOUD_PROVIDER")
    id = f"osdu/partition/{data_partition}/{policy_id}"

    sha1 = hashlib.sha1(contents.decode("utf-8").encode()).hexdigest()
    response.headers["X-SHA-1"] = sha1

    audit_msg = f"user: {auth_data.user_id} xuserid: {auth_data.x_user_id} sha1: {sha1} requestid: {x_request_id} traceid: {x_b3_traceid} spanid: {x_b3_spanid} parentspanid: {x_b3_parentspanid} forward: {x_forwarded_for}"

    validated = False
    # Verify if ENABLE_VERIFY_POLICY is set and not Mocking OPA
    if conf.ENABLE_VERIFY_POLICY and cloud_provider not in (None, conf.MOCK):
        validated = verify_policy_with_opa(
            policy_id=policy_id,
            auth_data=auth_data,
            correlation_id=context["correlation_id"],
            data=contents.decode("utf-8"),
        )

    if cloud_provider is None:
        # Help support MOCK testing
        logger.critical(
            f"Error: CLOUD_PROVIDER ENV VAR not set / Mocking results for put /policies/osdu/partition/{data_partition}/{id}"
        )
        result = OpaResponse()
        result.ok = True
        result.status_code = HTTP_200_OK
        result.message = f"MOCK Policy {id} added to OPA"
    elif cloud_provider == conf.MOCK:
        logger.critical(
            f"Warning: CLOUD_PROVIDER ENV VAR set to Mock results {cloud_provider} for put /policies/osdu/partition/{data_partition}/{id}"
        )
        result = OpaResponse()
        result.ok = True
        result.status_code = HTTP_200_OK
        result.message = f"MOCK Policy {id} added to OPA"
    elif cloud_provider == conf.LOCAL:
        # LOCAL Development ENV
        logger.critical(
            f"Error: CLOUD_PROVIDER ENV VAR LOCAL, Using local / direct OPA {conf.OPA_URL} put /policies/osdu/partition/{data_partition}/{id} sha1 {sha1}"
        )
        result = opa.put_opa_policy_direct(policy_id=id, data=contents)
        # result = opa.put_opa_policy_direct(policy_id=id, data=contents.decode('utf-8'))
    else:
        logger.info(f"bundle put policy_id: {policy_id} id: {id} {auth_data.user_id}")
        result = bundle.put_policy(
            data_partition=data_partition, policy_id=policy_id, policy=contents
        )

    if result.ok:
        # TODO dev testing - fix return
        logger.info(f"AUDIT put policy_id: {policy_id} id: {id} {audit_msg}")
        return {
            "policy_id": policy_id,
            "data_partition": data_partition,
            "opa_payload": contents,
            "status_code": result.status_code,
            "validated": validated,
            "sha1": sha1,
            "status": result.ok,
            "message": result.message,
        }
    else:
        if cloud_provider == conf.LOCAL:
            logger.error(f"Error when talking to OPA {result.message}")
            raise HTTPException(
                status_code=result.status_code,
                detail=f"Error when talking to OPA {result.message}",
            )
        else:
            logger.error(
                f"Error when talking to bundle service {result.status_code} {result.message}. {audit_msg}"
            )
            raise HTTPException(
                status_code=result.status_code,
                detail=f"Error when talking to bundle service {result.message}",
            )
