# from email import policy
import fastapi
from fastapi import HTTPException, Depends, Response
import json
import logging
import re
import opa
import conf
from starlette_context import context
import correlation
from pydantic import BaseModel

from starlette.status import (
    HTTP_500_INTERNAL_SERVER_ERROR,
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_406_NOT_ACCEPTABLE,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from models.translate import TranslateItem
from auth import auth
import translate.convert
from translate.translate_exceptions import PolicyDenyError, PolicyNotFoundError


class Detail(BaseModel):
    detail: str


router = fastapi.APIRouter()
logger = logging.getLogger(__name__)


@router.post(
    "/translate",
    tags=["translate"],
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_404_NOT_FOUND: {"model": Detail},
        HTTP_406_NOT_ACCEPTABLE: {"model": Detail},
        HTTP_422_UNPROCESSABLE_ENTITY: {"model": Detail},
        HTTP_500_INTERNAL_SERVER_ERROR: {"model": Detail},
    },
)
def translate_policy_api(
    response: Response,
    trans: TranslateItem,
    auth_data: auth.Auth = Depends(auth.require_authorized_user),
):
    """
    ## Translate policy
    Given an OPA query that should be partially evaluated, return an ElasticSearch request body

    In the body of the request the JSON schema should match "TranslateItem".
    Please note: xuserid, token and datapartitionid are now actively inserted into input request

    ### Permission required to use this API:

    `policy.service.user` or `policy.service.admin`
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    posted_data = json.loads(trans.tojson())  # convert to dict for easy manipulation
    logger.debug(f"data: {posted_data}")

    if "input" not in posted_data:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail="Input not found in data"
        )

    posted_data["input"]["token"] = auth_data.access_token
    posted_data["input"]["xuserid"] = auth_data.x_user_id
    posted_data["input"]["datapartitionid"] = auth_data.data_partition_id
    # logger.debug(f"data after adding auth: {posted_data}")

    logger.debug(f"ENABLE_TRANSLATE_PREPROCESSOR {conf.ENABLE_TRANSLATE_PREPROCESSOR}")
    if conf.ENABLE_TRANSLATE_PREPROCESSOR:
        return translate_preprocess(
            posted_data=posted_data, data_partition=auth_data.data_partition_id
        )
    else:
        return translate_policy(
            posted_data=posted_data, data_partition=auth_data.data_partition_id
        )


def translate_policy(posted_data, data_partition):
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    logger.debug(f"compile {posted_data}")
    opa_response = opa.compile(
        query=json.dumps(posted_data), name="compile api for translate"
    )

    if opa_response.ok:
        js = opa_response.json
        logger.debug(f"OPA compile response: {js}")
        try:
            if "queries" in opa_response.text:
                opa_ast = opa_response.json["result"]["queries"]
                logger.debug(f"OK opa response {opa_ast}")
            else:
                if policy_exists(query=posted_data, data_partition=data_partition):
                    logger.warning(f"No queries in opa response {js} but policy exists")
                    # When not getting a valid response from OPA, we want to still make search happy
                    # If policy exists return a false match response
                    # If policy does not exist return a HTTP_404_NOT_FOUND
                    return {"query": {"match_none": {}}}
                else:
                    logger.error(
                        f"No queries in opa response {js} and policy does not exists"
                    )
                    raise PolicyNotFoundError(
                        f"No query in opa response. Translate failed. Policy may not exist. {js}"
                    )

            es_request_body = translate.convert.convert_opa_ast_json_to_es_request_body(
                # opa_ast, js
                opa_ast,
                json.dumps(posted_data["input"]),
            )
            return es_request_body
        except PolicyNotFoundError as err:
            logger.error(err)
            raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail=f"{err}")
        except PolicyDenyError as err:
            logger.error(err)
            raise HTTPException(
                status_code=HTTP_406_NOT_ACCEPTABLE,
                detail=f"An error occurred when talking translate service. {err}",
            )
            # return {
            #    "query": {
            #        "must_not": {}
            #        }
            #    }
        except NotImplementedError as err:
            logger.error(err)
            raise HTTPException(
                status_code=HTTP_406_NOT_ACCEPTABLE,
                detail=f"An error occurred when talking translate service. {err}",
            )
        except KeyError as err:
            logger.error(err)
            raise HTTPException(
                status_code=HTTP_406_NOT_ACCEPTABLE,
                detail=f"An error occurred when talking translate service. {err}",
            )
        except Exception as err:
            logger.error(err)
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(err).__name__, err.args)
            logger.error(message)
            raise HTTPException(
                status_code=HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"An error occurred when talking translate service. {err} {message}",
            )
    else:
        logger.error(f"Error from OPA: {opa_response.message}")
        raise HTTPException(
            status_code=opa_response.status_code,
            detail=f"OPA Error: {opa_response.message}",
        )


def policy_exists(query, data_partition):
    """
    If policy exists
    """
    if "query" in query:
        p = query["query"]
        if p.startswith("data.osdu.partition["):
            policy_id = p.split(".")[3] + ".rego"
            opa_response = opa.fetch_opa_policy_direct(
                conf.PARTITION_BUNDLE_ROOT.format(data_partition)
                + "/{0}".format(policy_id),
                name="fetch opa policy direct for policy exist",
            )
        elif p.startswith("data.osdu.instance."):
            policy_id = p.split(".")[3] + ".rego"
            opa_response = opa.fetch_opa_policy_direct(
                conf.INSTANCE_BUNDLE_ROOT + "/{0}".format(policy_id)
            )
        else:
            logger.info(f"policy_exists unexpected query {p}")
            return False
        return opa_response.ok


def translate_preprocess(posted_data, data_partition):
    """
    Translate Preprocessor

    Step 1
    Call OPA DataAPI with the original "input" to retrieve the config with preprocessor results
    and add the preprocessor results to the original input.

    Step 2
    If preprocess_config.has_allow_rule == true, do translate (compile/translate) with allow query:
    """
    logger.debug(f"type {type(posted_data)}")
    logger.debug(f"posted_data {posted_data}")

    # Determine path
    if "query" in posted_data:
        original_query = posted_data["query"]
        if original_query.startswith("data.osdu.partition["):
            # "query": "data.osdu.partition[\"{{partition_id"}}\"].search.allow == true"
            # policy_id = p.split('.')[3] + ".rego"
            policy_id = original_query.split(".")[3]
            # rego_package_name + "/" + rego_rule_name
        elif original_query.startswith("data.osdu.partition."):
            # "query": "data.osdu.partition.{{partition_id}}.search.allow == true"
            policy_id = original_query.split(".")[4]
        elif original_query.startswith("data.osdu.instance."):
            policy_id = original_query.split(".")[3]
        else:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail="Translate API Error: expected query in format data.osdu.partition...",
            )
        path = conf.PARTITION_BUNDLE_ROOT.format(data_partition) + "/{0}".format(
            policy_id
        )
    else:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Translate API Error: expected query in request",
        )
    logger.debug(f"PolicyID: {policy_id} DATA_PARTITION: {data_partition} PATH: {path}")

    opa_response = opa.data(
        query=json.dumps(posted_data),
        path=path,
        name="data api for preprocess translate",
    )

    # simple string search, should change this later
    allow_rule = '"has_allow_rule":true'
    deny_rule = '"has_deny_rule":true'
    found_allow = False
    found_deny = False

    if opa_response.ok:
        logger.debug(f"Opa Response: {opa_response.json}")
        logger.debug(f"Opa Response text: {opa_response.text}")
        # posted_data = posted_data + opa_response.json["result"]["queries"]
        logger.debug(f"New input: {posted_data}")

        if "input_from_preprocessor" in opa_response.text:
            input_from_preprocessor = opa_response.json["result"]["preprocess_config"][
                "input_from_preprocessor"
            ]
            logger.debug(
                f"input_from_preprocessor: {input_from_preprocessor} {type(input_from_preprocessor)}"
            )
            for attribute in input_from_preprocessor:
                logger.debug(f"attribute: {attribute}")
                posted_data["input"][attribute] = input_from_preprocessor[attribute]
            logger.debug(
                f"revised posted_data with input_from_preprocessor: {posted_data}"
            )

        if "es_subquery" in opa_response.text:
            logger.debug("es_subquery exists in policy")
            es_subquery = opa_response.json["result"]["preprocess_config"][
                "es_subquery"
            ]
            logger.debug(f"es subquery to be returned: {es_subquery}")
            return es_subquery

        if "==" in original_query:
            logger.debug(f"Query already has == {original_query} for path {path}")
            # remove everything after last .
            cleaned_query = original_query.rsplit(".", 1)[0]
            logger.debug(f"split {cleaned_query}")
        else:
            cleaned_query = original_query

        if re.search(allow_rule, opa_response.text):
            found_allow = True
            logger.debug(f"old query: {original_query}")
            posted_data["query"] = cleaned_query + ".allow == true"

            # for debug
            query = posted_data["query"]
            logger.debug(f"new allow query: {query}")
            logger.debug(f"Found allow rule: {allow_rule} {posted_data}")

            allow_es_request_body = translate_policy(
                posted_data=posted_data, data_partition=data_partition
            )
            logger.debug(f"allow = {allow_es_request_body}")

        if re.search(deny_rule, opa_response.text):
            found_deny = True
            posted_data["query"] = f"{cleaned_query}.deny == true"

            query = posted_data["query"]
            logger.debug(f"new deny query: {query}")
            logger.debug(f"Found deny rule {deny_rule} {posted_data}")

            deny_es_request_body = translate_policy(
                posted_data=posted_data, data_partition=data_partition
            )
            logger.debug(f"unwrapped deny = {deny_es_request_body}")
            wrapped_deny_eq_request_body = wrap_deny(deny_es_request_body)
            logger.debug(f"wrapped deny = {wrapped_deny_eq_request_body}")

        if found_deny and found_allow:
            logger.debug("found both allow and deny")
            return wrap_combine(
                allow_es_body=allow_es_request_body,
                deny_es_body=wrapped_deny_eq_request_body,
            )
        elif found_deny:
            # just deny
            logger.debug("Found only deny")
            return wrapped_deny_eq_request_body
        elif found_allow:
            # just allow
            logger.debug("Found only allow")
            return allow_es_request_body
        else:
            # no allow or deny rules
            logger.debug("Found no allow or deny rules")
            return translate_policy(
                posted_data=posted_data, data_partition=data_partition
            )
    else:
        logger.error(f"Error from data OPA: {opa_response.message}")
        raise HTTPException(
            status_code=opa_response.status_code,
            detail=f"OPA Error: {opa_response.message}",
        )


def wrap_combine(allow_es_body, deny_es_body):
    """
    Wrap combination of allow and deny to make es happy
    {
        "bool": {
            "must": [
                allow,
                deny
            ]
        }
    }
    """
    logger.debug(f"allow_es_body before wrap: {allow_es_body}")
    logger.debug(f"deny_es_body before wrap: {deny_es_body}")
    wrap = {
        "query": {"bool": {"must": [allow_es_body["query"], deny_es_body["query"]]}}
    }
    logger.debug(f"combine wrap {wrap}")
    return wrap


def wrap_deny(deny_es_body):
    """
    Wrap deny to make es happy
    {
        "bool": {
            "must_not": [
                {
                }
            ]
        }
    }
    """
    wrap = {"query": {"bool": {"must_not": [deny_es_body["query"]]}}}
    logger.debug(f"deny wrap {wrap}")
    return wrap
