# Copyright © Amazon Web Services
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Tuple
from osdu_api.providers.blob_storage import get_client
from osdu_api.providers.types import FileLikeObject
import boto3

from bundles.storage import BundleStorageClient
import conf

logger = logging.getLogger(__name__)
logger.setLevel(conf.LOG_LEVEL)

class FileNotFound(Exception):
    def __init__(self, filename: str):
        self.message = f'ERROR: File {filename} was not found.'
        super().__init__(self.message)

class AWSBundleStorageClient(BundleStorageClient):
    def __init__(self) -> None:
        self.client = get_client()
        self.bucket_name = os.environ["POLICY_BUCKET"]
        self.content_type = "application/x-gtar"
        session = boto3.session.Session()
        self.s3_client =session.client('s3', region_name="us-east-1")
        logger.debug(f"bucket: {self.bucket_name}")


    def _get_bucket_uri(self, filename: str) -> str:
        return f"s3://{self.bucket_name}/{filename}"

    def download_file(self, filename: str, file: FileLikeObject) -> Tuple[FileLikeObject, str]:
        try:
            uri = self._get_bucket_uri(filename)

            if self._does_file_exist(uri):
                self._download_to_file(uri, file)
                return file, uri
            else:
                logger.info(f"{filename} does not exist")
                raise FileNotFound(filename)
            
        except Exception as e:
            logger.error(f"Failed to download file from {uri}: {e}")
        return None, None

    def upload_file(self, name: str, file: FileLikeObject) -> str:
        try:
            uri = self._get_bucket_uri(name)
            self.client.upload_file(uri, file, self.content_type)
            return uri
        except Exception as e:
            logger.error(f"Failed to upload file to {uri}: {e}")

    # TODO: Fix bug in AWS python SDK and replace
    def _does_file_exist(self, uri: str) -> bool:
            """Verify if a file exists in the given URI.

            :param uri: The AWS URI of the file.
            :type uri: str
            :return: A boolean indicating if the file exists
            :rtype: bool
            """

            # assuming the URI here is an s3:// URI
            # get the bucket name and path to object
            bucket_name, object_name = self._split_s3_path(uri)
            try:
                # try to get the s3 metadata for the object, which is a 
                # fast operation no matter the size of the data object
                self.s3_client.head_object(Bucket=bucket_name, Key=object_name)
            except Exception as e:
                logger.info(f"{uri} does not exist {e}")
                return False
            return True

    def _split_s3_path(self, s3_path:str):
        """split a s3:// path into bucket and key parts

        Args:
            s3_path (str): an s3:// uri

        Returns:
            tuple: bucket name, key name ( with path )
        """
        path_parts=s3_path.replace("s3://","").split("/")
        bucket=path_parts.pop(0)
        key="/".join(path_parts)
        return bucket, key

    # TODO: Fix bug in AWS python SDK and replace
    def _download_to_file(self, uri: str, file: FileLikeObject) -> None:
        """Download file from the given URI.

        :param uri: The AWS URI of the file.
        :type uri: str
        :param file: The file where to download the blob content
        :type file: FileLikeObject
        """
        # assuming the URI here is an s3:// URI
        # get the bucket name, path to object
        bucket_name, object_name = self._split_s3_path(uri)
        self.s3_client.download_fileobj(bucket_name, object_name, file)
